/* This file is Copyright 2002 Level Control Systems.  See the included LICENSE.txt file for details. */

/* 
   This library was downloaded from: http://www.mike95.com 
 
   This library is copyright.  It may freely be used for personal purposes  
   if the restriction listed below is adhered to. 
       Author: Michael Olivero 
       Email:  mike95@mike95.com 
 
   //=============================== 
   //Start of Restriction Definition 
   //=============================== 
   Anyone can have full use of the library provided they keep this complete comment 
   with the source.  Also I would like to ask if any changes are made to the 
   code for efficiency reasons, please let me know so I may look into your change and 
   likewise incorporate it into library.  If a suggestion makes it into the library, 
   your credits will be added to this information. 
 
   Authors of Computer related books are welcome to include this source code as part 
   of their publishing, provided credit the Author and make note of where the source 
   code was obtained from: http://www.mike95.com 
   //============================= 
   //End of Restriction Definition 
   //============================= 
 
 
   Description: 
   Visit http://www.mike95.com/c_plusplus/classes/JString/ 

   This library avoids the need to use pointers to char or manual
   memory manipulation while working with Strings.  This class has been
   based on the Java String class, and thus has all of the Java public functionality 
   with a few extras [replace()] needed for useful functionality.
 
   //The following people have contributed to the solution 
   //of bugs in this library 
   //===================================================== 
   //Carl Pupa, email: pumacat@erols.com 
*/
 
#ifndef MuscleString_h
#define MuscleString_h

#include <string.h> 
#include <ctype.h> 
#include "support/Flattenable.h"

namespace muscle {

#ifndef SMALL_MUSCLE_STRING_LENGTH
# define SMALL_MUSCLE_STRING_LENGTH 7  // strings shorter than this length can be stored inline, without requiring an extra new[].
#endif

class String;

/** A nice hashing function for use with Strings. */
uint32 StringHashFunc(const String &);

/** A nice hashing function for use with (const char *)'s */
uint32 CStringHashFunc(const char * const &);

/** A function for comparing (const char *)'s -- calls strcmp() */
int CStringCompareFunc(const char * const &, const char * const &, void *);

/** A character string class, similar to Java's java.lang.String */
class String : public Flattenable {
public:
   /** Constructor.
    *  @param str If non-NULL, the initial value for this String.
    */
   String(const char * str = NULL);

   /** Copy Constructor. */
   String(const String & str);

   /** Destructor. */
   virtual ~String() {if (_buffer != _smallBuffer) delete [] _buffer;}

   /** Assignment Operator. */
   const String &   operator = (char val);

   /** Assignment Operator. */
   const String &   operator = (const char * val);

   /** Assignment Operator. */
   const String &   operator = (const String &Rhs);

   /** Append Operator. 
    *  @param Rhs A string to append to this string.
    */
   const String &   operator +=(const String &Rhs);

   /** Append Operator.
    *  @param ch A character to append to this string.
    */
   const String &   operator +=(const char ch);
   
   /** Remove Operator. 
    *  @param Rhs A substring to remove from this string;  the
    *             last instance of the substring will be cut out.
    *             If (Rhs) is not found, there is no effect.
    */
   const String &   operator -=(const String &Rhs);

   /** Remove Operator.
    *  @param ch A character to remove from this string;  the last
    *            instance of this char will be cut out.  If (ch) is
    *            not found, there is no effect.
    */
   const String &   operator -=(const char ch);
   
   /** Append 'Stream' Operator.
    *  @param Rhs A String to append to this string.
    *  @return a non const String refrence to 'this' so you can chain appends.
    */
   String &   operator <<(const String& Rhs);   

   /** Append 'Stream' Operator.
    *  @param Rhs A const char* to append to this string.
    *  @return a non const String refrence to 'this' so you can chain appends.
    */
   String &   operator <<(const char* Rhs);   
   
   /** Append 'Stream' Operator.
    *  @param Rhs An int to append to this string.
    *  @return a non const String refrence to 'this' so you can chain appends.
    */
   String &   operator <<(int Rhs);   

   /** Append 'Stream' Operator.
    *  @param Rhs A float to append to this string. Formatting is set at 2 decimals of precision.
    *  @return a non const String refrence to 'this' so you can chain appends.
    */
   String &   operator <<(float Rhs);   

   /** Append 'Stream' Operator.
    *  @param Rhs A bool to append to this string. Converts to 'true' and 'false' strings appropriately.
    *  @return a non const String refrence to 'this' so you can chain appends.
    */
   String &   operator <<(bool Rhs);

   /** Comparison Operator.  Returns true if the two strings are equal (as determined by strcmp()) */
   int            operator ==(const String &Rhs) const;

   /** Comparison Operator.  Returns true if the two strings are not equal (as determined by strcmp()) */
   int            operator !=(const String &Rhs) const;

   /** Comparison Operator.  Returns true if this string comes before (Rhs) lexically. */
   int            operator < (const String &Rhs) const;

   /** Comparison Operator.  Returns true if this string comes after (Rhs) lexically. */
   int            operator > (const String &Rhs) const;

   /** Comparison Operator.  Returns true if the two strings are equal, or this string comes before (Rhs) lexically. */
   int            operator <=(const String &Rhs) const;

   /** Comparison Operator.  Returns true if the two strings are equal, or this string comes after (Rhs) lexically. */
   int            operator >=(const String &Rhs) const;

   /** Array Operator.  Used to get easy access to the characters that make up this string.
    *  @param Index Index of the character to return.  Be sure to only use valid indices!
    */
   char           operator [](uint32 Index) const;

   /** Array Operator.  Used to get easy access to the characters that make up this string.
    *  @param Index Index of the character to set.  Be sure to only use valid indices!
    */
   char&          operator [](uint32 Index);

   /** Returns the character at the (index)'th position in the string.
    *  @param index A value between 0 and (Length()-1), inclusive.
    *  @return A character value.
    */
   char           CharAt(uint32 index) const;
 
   /** Compares this string to another string using strcmp() */
   int            CompareTo(const String &anotherString) const;

   /** Returns a C-style pointer to our held character string. */
   const char*    Cstr() const {return _buffer ? _buffer : "";}

   /** Convenience synonym for Cstr(). */
   const char * operator()() const {return Cstr();}  

   /** Returns true iff this string ends with (suffix) */
   bool           EndsWith(const String &suffix) const;

   /** Returns true iff this string is equal to (string), as determined by strcmp(). */
   bool           Equals(const String &string) const;

   /** Returns the first index of (ch) in this string, or -1 if not found. */
   int            IndexOf(char ch) const;

   /** Returns the first index of (ch) in this string starting at or after (fromIndex), or -1 if not found. */
   int            IndexOf(char ch, uint32 fromIndex) const;

   /** Returns the first index of substring (str) in this string, or -1 if not found. */
   int            IndexOf(const String &str) const;

   /** Returns the first index of substring (str) in this string starting at or after (fromIndex), or -1 if not found. */
   int            IndexOf(const String &str, uint32 fromIndex) const;

   /** Returns the last index of (ch) in this string. */
   int            LastIndexOf(char ch) const;

   /** Returns the last index of (ch) in this string starting at or after (fromIndex), or -1 if not found. */
   int            LastIndexOf(char ch, uint32 fromIndex) const;

   /** Returns the last index of substring (str) in this string, or -1 if not found. */
   int            LastIndexOf(const String &str) const;

   /** Returns the last index of substring (str) in this string starting at or after (fromIndex), or -1 if not found. */
   int            LastIndexOf(const String &str, uint32 fromIndex) const;

   /** Returns the number of characters in the string (not including the terminating NUL byte) */
   const uint32   Length() const { return _length; }

   /** Returns true iff this string starts with (prefix) */
   bool           StartsWith(const String &prefix) const;

   /** Returns true iff this string starts with the first (tooffset) characters of (prefix) */
   bool           StartsWith(const String &prefix, uint32 toffset) const; 

   /** Returns a string that consists of (str) prepended to this string.  Does not modify the String it is called on. */
   String         Prepend(const String & str) const;

   /** Returns a string that consists of (str) appended to this string.  Does not modify the String it is called on. */
   String         Append(const String & str) const;

   /** Returns a string that consists of only the last part of this string, starting with index (beginIndex).  Does not modify the string it is called on*/
   String         Substring(uint32 beginIndex) const; 

   /** Returns a string that consists of only the characters in this string from range (beginIndex) to (endIndex-1).  Does not modify the string it is called on*/
   String         Substring(uint32 beginIndex, uint32 endIndex) const; 

   /** Returns an all lower-case version of this string.  Does not modify the string it is called on. */
   String         ToLowerCase() const; 

   /** Returns an all upper-case version of this string.  Does not modify the string it is called on. */
   String         ToUpperCase() const; 

   /** Returns an version of this string that has all leading and trailing whitespace removed.  Does not modify the string it is called on. */
   String         Trim() const;  

   /** Like CompareTo(), but case insensitive. */
   int            CompareToIgnoreCase(const String &s) const             {return ToLowerCase().CompareTo(s.ToLowerCase());}

   /** Like EndsWith(), but case insensitive. */
   bool           EndsWithIgnoreCase(const String &s) const              {return ToLowerCase().EndsWith(s.ToLowerCase());}

   /** Like Equals(), but case insensitive. */
   bool           EqualsIgnoreCase(const String &s) const                {return ToLowerCase().Equals(s.ToLowerCase());}

   /** Like IndexOf(), but case insensitive. */
   int            IndexOfIgnoreCase(const String &s) const               {return ToLowerCase().IndexOf(s.ToLowerCase());}

   /** Like IndexOf(), but case insensitive. */
   int            IndexOfIgnoreCase(const String &s, uint32 f) const     {return ToLowerCase().IndexOf(s.ToLowerCase(),f);}

   /** Like IndexOf(), but case insensitive. */
   int            IndexOfIgnoreCase(char ch) const                               {return ToLowerCase().IndexOf((char)tolower(ch));}

   /** Like IndexOf(), but case insensitive. */
   int            IndexOfIgnoreCase(char ch, uint32 f) const                     {return ToLowerCase().IndexOf((char)tolower(ch),f);}

   /** Like LastIndexOf(), but case insensitive. */
   int            LastIndexOfIgnoreCase(const String &s) const           {return ToLowerCase().LastIndexOf(s.ToLowerCase());}

   /** Like LastIndexOf(), but case insensitive. */
   int            LastIndexOfIgnoreCase(const String &s, uint32 f) const {return ToLowerCase().LastIndexOf(s.ToLowerCase(),f);}

   /** Like LastIndexOf(), but case insensitive. */
   int            LastIndexOfIgnoreCase(char ch) const                           {return ToLowerCase().LastIndexOf((char)tolower(ch));}

   /** Like LastIndexOf(), but case insensitive. */
   int            LastIndexOfIgnoreCase(char ch, uint32 f) const                 {return ToLowerCase().LastIndexOf((char)tolower(ch),f);}

   /** Like StartsWith(), but case insensitive. */
   bool           StartsWithIgnoreCase(const String &s) const            {return ToLowerCase().StartsWith(s.ToLowerCase());}

   /** Like StartsWith(), but case insensitive. */
   bool           StartsWithIgnoreCase(const String &s, uint32 o) const  {return ToLowerCase().StartsWith(s.ToLowerCase(),o);}

   /** Returns a hash code for this string */
   uint32 HashCode() const;

   /** Replaces all instances of (oldChar) in this string with (newChar) */
   void         Replace(char oldChar, char newChar); 

   /** Replaces all instances of (match) in this string with (replace) */
   void         Replace(const String& match, const String& replace); 
 
   /** Part of the Flattenable interface.
    *  @return false
    */
   virtual bool            IsFixedSize() const {return false;}

   /** Part of the Flattenable interface.
    *  @return B_STRING_TYPE
    */
   virtual type_code       TypeCode() const {return B_STRING_TYPE;}

   /** Part of the Flattenable interface.
    *  @return Length()+1  (the +1 is for the terminating NUL byte)
    */
   virtual uint32          FlattenedSize() const;

   /** Part of the Flattenable interface.  Flattens our string into (buffer).
    *  @param buffer A byte array to receive the flattened version of this string.
    *                There must be at least FlattenedSize() bytes available in this array.
    *  The clever secret here is that a flattened String is just a C-style
    *  null-terminated character array, and can be used interchangably as such.
    */
   virtual void            Flatten(uint8 *buffer) const;

   /** Unflattens a String from (buf).
    *  @param buf an array of (size) bytes.
    *  @size the number of bytes in (buf).
    *  @return B_NO_ERROR (never fails!)
    */
   virtual status_t        Unflatten(const uint8 *buf, uint32 size);

   /** Makes sure that we have pre-allocated enough space for a string at least (numChars)
    *  ASCII characters long.  If not, this method will try to allocate the space.
    *  @param numChars How much space to pre-allocate, in ASCII characters.
    *  @returns B_NO_ERROR on success, or B_ERROR on failure (out of memory).
    */ 
   status_t Prealloc(uint32 newBufLen) {return EnsureBufferSize(newBufLen+1, true);}

private:
   status_t EnsureBufferSize(uint32 newBufLen, bool retainValue);
   char _smallBuffer[SMALL_MUSCLE_STRING_LENGTH+1];  // store very small strings in-line, to avoid dynamic allocation
   char * _buffer;            // Stores the chars.  May point at (_smallBuffer), or a dynamically allocated buffer
   uint32 _bufferLen;         // Number of bytes pointed to by (_buffer)
   uint32 _length;            // cached strlen(_buffer)

   void verifyIndex(uint32 number) const;
};

// This method tries to ensure that at least (newBufLen) chars
// are available for storing data in.  (requestedBufLen) should include
// the terminating NUL.  If (retainValue) is true, the current string value
// will be retained; otherwise it should be set right after this call
// returns...
inline status_t
String::EnsureBufferSize(uint32 requestedBufLen, bool retainValue)
{
   if ((requestedBufLen > 0)&&((_buffer == NULL)||(requestedBufLen > _bufferLen)))
   {
      // If we're doing an initial allocation, just allocate the bytes requested.
      // If it's a re-allocation, allocate more than requested as it's more likely
      // to happen yet another time...
      uint32 newBufLen = (_buffer == NULL) ? requestedBufLen : (requestedBufLen * 2);
      char * newBuf = (requestedBufLen <= sizeof(_smallBuffer)) ? _smallBuffer : newnothrow char[newBufLen]; 
      if (newBuf == _smallBuffer) newBufLen = sizeof(_smallBuffer);
      if (newBuf)
      {
         if ((retainValue)&&(_buffer)&&(newBuf != _buffer)) strncpy(newBuf, _buffer, newBufLen);
         if (_buffer != _smallBuffer) delete [] _buffer;
         _buffer = newBuf;
         _bufferLen = newBufLen;
      }
      else 
      {
         WARN_OUT_OF_MEMORY;
         return B_ERROR;
      }
   }
   return B_NO_ERROR;
}

inline void 
String::verifyIndex(uint32 index) const 
{ 
   MASSERT(index < _length, "Index Out Of Bounds Exception"); 
} 

//Friend Functions
//================

inline const String
operator+(const String & Lhs, const String &Rhs) 
{
   String ret(Lhs);
   ret += Rhs;
   return ret; 
}



inline const String
operator+(const String & Lhs, const char *Rhs) 
{
   String ret(Lhs);
   ret += Rhs;
   return ret; 
}


inline const String
operator+(const char * Lhs, const String & Rhs) 
{
   String ret(Lhs);
   ret += Rhs;
   return ret; 
}



inline const String
operator+(const String & Lhs, char Rhs) 
{
   String ret(Lhs);
   ret += Rhs;
   return ret; 
}


inline const String
operator-(const String & Lhs, const String &Rhs) 
{
   String ret(Lhs);
   ret -= Rhs;
   return ret; 
}


inline const String
operator-(const String & Lhs, const char *Rhs) 
{
   String ret(Lhs);
   ret -= Rhs;
   return ret; 
}


inline const String
operator-(const char *Lhs, const String &Rhs) 
{
   String ret(Lhs);
   ret -= Rhs;
   return ret; 
}


inline const String
operator-(const String & Lhs, char Rhs) 
{
   String ret(Lhs);
   ret -= Rhs;
   return ret; 
}

};  // end namespace muscle

#endif
